import logging

from django.core.paginator import Paginator
from django.db.models import Q
from django.http import HttpResponseForbidden, JsonResponse
from django.shortcuts import redirect
from django.utils.text import slugify
from django.views import View

logger = logging.getLogger(__name__)

class Select2View(View):
    RESULTS = "results"
    TEXT = "text"
    CHILDREN = "children"
    IMAGE = "img"
    ID = "id"

    TERM_QS = "term"
    Q_QS = "q"
    _TYPE_QS = "_type"
    PAGE_QS = "page"

    PAGINATION_SIZE = 20
    PAGINATION = "pagination"
    MORE = "more"

    SEARCH_PARAMS = [] # Need to override
    SEARCH_TYPE = "" # Need to override
    ORDER_BY_PARAMS = [] # Need to override
    MODEL = None # Need to override
    MODEL_DEFAULT_MANAGER = "_default_manager"
    MANAGER = MODEL_DEFAULT_MANAGER
    MODEL_VERBOSE_NAMES = {} # Need to override

    def get(self, request, *args, **kwargs):
        self.term = request.GET.get(self.TERM_QS, "")
        try:
            self.page = int(request.GET.get(self.PAGE_QS, 1))
        except Exception as e:
            self.page = 1

        default_manager = getattr(self.MODEL, self.MODEL_DEFAULT_MANAGER, None)
        model_manager = getattr(self.MODEL, self.MANAGER, default_manager)
        if not model_manager:
            model_manager = default_manager

        elements = model_manager.all()
        elements = self.get_initial_filter(elements, request, *args, **kwargs)
        if self.term:
            elements = self.filter_by_term(elements)

        group_by_context = self.group_elements_by_context(elements)

        has_more = []
        results = []
        for key in group_by_context:
            children = []
            paginator = Paginator(group_by_context[key], self.PAGINATION_SIZE)
            page = paginator.page(self.page)
            has_more.append(page.has_next())
            for element in page.object_list:
                element_context = {
                    self.ID : element.pk
                }
                str_element = self.stringify_element(element)
                if str_element:
                    element_context[self.TEXT] = str_element

                element_context.update(self.get_extra_context(element))
                children.append(element_context)
            results.append({
                self.TEXT : key,
                self.CHILDREN : children
            })

        response_context = {
            self.RESULTS : results,
            self.PAGINATION: {
                self.MORE: any(has_more)
            }
        }
        return JsonResponse(data=response_context)

    def get_extra_context(self, element):
        return {}

    def stringify_element(self, element):
        if element:
            return str(element)
        else:
            return None

    def group_elements_by_context(self, elements):
        group_by_context = {}
        for element in elements:
            model_name = element.__class__.__name__
            verbose_name = self.MODEL_VERBOSE_NAMES.get(model_name, model_name)
            if verbose_name in group_by_context:
                group_by_context[verbose_name].append(element)
            else:
                group_by_context[verbose_name] = [element]
        return group_by_context

    def filter_by_term(self, elements):
        qs = None
        filter_context = dict()
        for param in self.SEARCH_PARAMS:
            filter_context = {
                param + "__" + self.SEARCH_TYPE : self.term
            }
            if not qs:
                qs = Q(**filter_context)
                continue

            qs |= Q(**filter_context)
        return elements.filter(qs).order_by(*self.ORDER_BY_PARAMS)

    def get_initial_filter(self, queryset, request, *args, **kwargs):
        return queryset

class SlugifyView(View):
    MODEL = None

    def get(self, request, *args, **kwargs):
        title = request.GET.get("title", "")
        pk = request.GET.get("pk", None)
        pk = pk if pk else None
        return JsonResponse(data={
            "slug" : self.get_unique_slug(title=title, pk=pk)
        })

    def get_unique_slug(title=None, pk=None):
        logger.debug("Transforming {} into a slug...".format(title))

        assert (title is not None and pk is not None)

        pk = pk if not bool(instance) else instance.pk
        title = title if not bool(instance) else instance.title
        slug = slugify(title)

        queryset = self.MODEL._default_manager.filter(~Q(pk=pk), slug=slug)
        if queryset:
            if pk:
                identifier = pk
            else:
                identifier = queryset.last().pk + 1 if queryset.exists() else 1
            slug = "{}-{}".format(slug, identifier)
        logger.debug("{} into a slug -> {}".format(title, slug))
        return slug

class AjaxSlugifyView(SlugifyView):
    def has_authorization(self, request, *args, **kwargs):
        return True

    def dispatch(self, request, *args, **kwargs):
        if request.is_ajax():
            if self.has_authorization(request, *args, **kwargs):
                return super().dispatch(request, *args, **kwargs)
            else:
                return HttpResponseForbidden("Not enough permissions to request slug via ajax")
        else:
            return HttpResponseBadRequest

class AjaxView(View):
    def dispatch(self, request, *args, **kwargs):
        if self.request.is_ajax():
            return super().dispatch(request, *args, **kwargs)
        else:
            return HttpResponseBadRequest()

class GenericDeleteView(View):
    model = None

    def has_authorization(self, request, instance, *args, **kwargs):
        return True

    def post(self, request, *args, **kwargs):
        pk = kwargs.pop("pk")
        redirect_url = request.POST.get("redirect")

        instance = self.model._default_manager.get(pk=pk)
        if not self.has_authorization(request, instance, *args, **kwargs):
            return HttpResponseForbidden("No authorization to delete")

        self.process_delete(instance)

        response_context = {
            "result" : "OK"
        }
        extra_response_context = self.get_extra_response_context()
        response_context.update(extra_response_context)

        if redirect_url:
            self.add_message(request)
            return redirect(redirect_url)
        else:
            return JsonResponse(
                response_context,
                safe=False
            )

    def process_delete(self, instance):
        instance.delete()

    def add_message(self, request):
        pass

    def get_extra_response_context(self):
        return {}



from django import template

register = template.Library()

@register.assignment_tag
def format_number(value,limit):
    array_number = []
    for digit in str(value):
        array_number.append(digit)
    formated = ""
    counter = 0
    for i in array_number:
        formated += str(i)
        counter += 1
        if counter == limit:
            formated += " "
            counter = 0
    return formated



from functools import wraps
from time import perf_counter
from typing import Any, Callable, Optional, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])

def timer(prefix: Optional[str] = None, precision: int = 6) -> Callable[[F], F]:

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            nonlocal prefix
            prefix = prefix if prefix is not None else f"{func.__name__}: "
            start = perf_counter()
            result = func(*args, **kwargs)
            end = perf_counter()
            print(f"{prefix}{end - start:.{precision}f}s - args: {args} - kwargs: {kwargs}")
            return result
        return cast(F, wrapper)

    return decorator



import importlib
import re
import logging
import traceback
from django import forms
from django.contrib import messages
from django.shortcuts import reverse
from django.conf import settings
from django.contrib.sites.models import Site
from django.contrib.contenttypes.models import ContentType
from django.db.models import Manager, ManyToManyField
from django.template.loader import get_template, TemplateDoesNotExist, render_to_string
from urllib.parse import urlparse, urljoin, parse_qs, urlencode, urlsplit, urlunsplit
from copy import copy
from itertools import chain

logger = logging.getLogger(__name__)

def normalize_domain(domain):
    return re.sub('[^\w\-_ ]', '', domain)

def get_model_attr(instance, field, default=[]):
    try:
        value = getattr(instance, field, None)
    except Exception as e:
        logger.exception(e)
        value = None
    is_manager = isinstance(value, Manager)
    if is_manager:
        value = value.all()
    return value

def get_attr_by_string(class_path):
    path_fragments = class_path.split(".")
    module_path = ".".join(path_fragments[:-1])
    class_name = path_fragments[-1]
    module = importlib.import_module(module_path)
    _class = getattr(module, class_name, None)
    if _class:
        return _class
    else:
        raise Exception(
            "Couldn't find the class with the path {}".format(class_path))

def exists_template(template_name):
    try:
        get_template(template_name)
        return True
    except TemplateDoesNotExist as e:
        return False

def render_to_string_or_empty(template_name, context, request=None, using=None):
    logger.debug("TEMPLATE NAME -> {}".format(template_name))
    try:
        content = render_to_string(
            template_name, context=context, request=request, using=using
        )
    except Exception as e:
        logger.exception("Error in rendering the template -> {}".format(template_name))
        is_preview = context.get("is_preview")
        if is_preview:
            content = traceback.format_exc()
        else:
            content = ""
    return content

def instance_to_dot_path(instance):
    if isinstance(instance, type):
        return ".".join([instance.__module__, instance.__name__])
    else:
        return ".".join([instance.__class__.__module__, instance.__class__.__name__])

def errors_serialize(form):
    errors_context = dict()
    for key in form.errors:
        if key in form:
            errors_context[key] = {
                "id": form[key].auto_id,
                "name": form[key].html_name,
                "errors": form.errors[key]
            }
    return errors_context

def to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def get_facebook_graph_url():
    facebook_graph_base_url = settings.FACEBOOK_GRAPH_URL
    if facebook_graph_base_url.endswith('/'):
        facebook_graph_base_url += "/"
    return facebook_graph_base_url

def is_absolute(url):
    return bool(urlparse(url).netloc)

def get_field_form_from_model(instance, field_name):
    instance_class = instance.__class__

    class CustomModelForm(forms.ModelForm):
        class Meta:
            model = instance_class
            fields = (field_name,)
    form = CustomModelForm(instance=instance)
    return form.fields[field_name]

def to_valid_variable_name(value):
    # Remove invalid characters
    value = re.sub('[^0-9a-zA-Z_]', '_', value)

    # Remove leading characters until we find a letter or underscore
    value = re.sub('^[^a-zA-Z_]+', '', value)

    return value

def replace_keys_with_valid_variable_name(context):
    new_context = dict()
    keys = context.keys()
    for key in keys:
        new_context[to_valid_variable_name(key)] = context.get(key)
    return new_context

def get_content_type_preview_url(printable_object):
    if printable_object:
        return reverse("preview-content-type", kwargs={
            "content_type_pk" : ContentType.objects.get_for_model(printable_object).pk,
            "pk" : printable_object.pk
        })
    else:
        return None

def send_message(request, msg_level, msg, **kwargs):
    fail_silently = kwargs.pop("fail_silently", False)

    if fail_silently:
        try:
            messages.add_message(request, msg_level, msg, **kwargs)
        except Exception as e:
            logger.exception("Failed to add message to the request -> {}, probably request is None".format(request))
    else:
        messages.add_message(request, msg_level, msg, **kwargs)

def set_query_parameters(url, **extra_query_params):
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    for param in extra_query_params:
        query_params[param] = extra_query_params[param]

    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))

def has_cached_prefetch_field(model_instance, attr_name):
    attr_value = getattr(model_instance, attr_name, None)
    if attr_value and isinstance(attr_value, list):
        return True

    return bool(
        getattr(model_instance, "_prefetched_objects_cache", {}).get(attr_name)
    )

def has_cached_related_field(model_instance, model_field=None, attr_name=None):
    if model_field is None and attr_name is None:
        return False

    field_name = model_field.field.name if model_field else attr_name

    return bool(
        getattr(model_instance._state, "fields_cache", {}).get(field_name)
    )

def get_select_related_choices_by_model(model_class):
    opts = model_class._meta

    direct_choices = (f.name for f in opts.fields if f.is_relation)

    reverse_choices = (
        f.field.related_query_name()
        for f in opts.related_objects if f.field.unique
    )

    return list(
        chain(direct_choices, reverse_choices)
    )

def get_prefetch_related_choices_by_model(model_class):
    opts = model_class._meta

    reverse_choices = (
        f.field.related_query_name()
        for f in opts.related_objects if not f.field.unique
    )

    direct_choices = (f.name for f in opts.get_fields() if isinstance(f, ManyToManyField))

    return list(
        chain(
            reverse_choices,
            direct_choices
        )
    )



from django.template import Library, Node
from django.forms import ChoiceField, MultipleChoiceField, ModelChoiceField, ModelMultipleChoiceField
from django.forms.boundfield import BoundWidget
from collections.abc import Iterable

register = Library()

@register.simple_tag
def get_selected_options(form, field_name, *sort_clauses):
    bound_field = form[field_name]

    if not isinstance(bound_field.field, ChoiceField):
        return []

    value = bound_field.value()
    if not isinstance(value, Iterable) or isinstance(value, str):
        value = [value]

    if isinstance(bound_field.field, ModelChoiceField) or isinstance(bound_field.field, MultipleChoiceField):
        selected_queryset = bound_field.field.queryset.filter(pk__in=value)
        if sort_clauses:
            selected_queryset = selected_queryset.order_by(*sort_clauses)
        choices = [(instance.pk, str(instance)) for instance in selected_queryset]
    else:
        choices = filter(lambda x : x[0] in value, bound_field.field.widget.choices)
    bound_field.field.widget.choices = choices
    return bound_field.subwidgets







import uuid

from django.db import models
from django.utils.translation import ugettext_lazy as _


# Create your models here.
class TrackingCreateUpdateChangesModel(models.Model):
    created_at = models.DateTimeField(_("Criado em"), auto_now_add=True, blank=True, null=True)
    updated_at = models.DateTimeField(_("Modificado em"), auto_now=True, blank=True)

    class Meta:
        abstract = True

class TrackingHistoricalModel(models.Model):
    ip_address = models.GenericIPAddressField(
        verbose_name=_('IP address'),
        null=True,
        blank=True
    )

    user_agent = models.CharField(
        verbose_name=_("User agent"),
        null=True,
        blank=True,
        max_length=256
    )

    class Meta:
        abstract = True

class UUIDTrackingCreateUpdateChangesModel(TrackingCreateUpdateChangesModel):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True



from django.db import connection
from django.db import reset_queries
from functools import wraps

def sql_queries():
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            prefix = f"{func.__name__}: "

            reset_queries()
            result = func(*args, **kwargs)

            queries = connection.queries

            print(f"{prefix} queries:")

            for query in queries:
                print(f"SQL: {query['sql']}")
                print(f"Took: {query['time']}")
            return result
        return wrapper
    return decorator



from django.db import connection
from django.db import reset_queries
from functools import wraps

def num_queries():
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            prefix = f"{func.__name__}: "

            reset_queries()
            result = func(*args, **kwargs)
            total_queries = len(connection.queries)

            print(f"{prefix} executed {total_queries} queries")

            return result
        return wrapper
    return decorator



from django.utils.module_loading import import_string
from django.template import Library, Node
import logging

logger = logging.getLogger(__name__)

register = Library()

@register.simple_tag
def get_query(model_path, manager="_default_manager", *filter_args, **filter_kwargs):
    try:
        model = import_string(model_path)
        manager = getattr(model, manager ,model._default_manager)
        return manager.filter(*filter_args, **filter_kwargs)
    except Exception as e:
        logger.exception("Error in getting the query")
        return None



from django.template import Library, Node
from datetime import timedelta
import logging

logger = logging.getLogger(__name__)

register = Library()

@register.simple_tag
def add_time(date, **kwargs):
    return date + timedelta(**kwargs)

@register.simple_tag
def subtract_time(date, **kwargs):
    return date - timedelta(**kwargs)



from django.template import Library, Node
from django.urls import resolve
from django.contrib.contenttypes.models import ContentType
from django.conf import settings
from django.utils import timezone
from django.core import serializers
from django.urls.exceptions import Resolver404
from urllib.parse import urlencode, urlparse
from ..utilities import render_to_string_or_empty
from datetime import datetime, date
import json
import logging
import requests

logger = logging.getLogger(__name__)

register = Library()

@register.simple_tag
def get_content_type(instance):
    return ContentType.objects.get_for_model(instance)

@register.simple_tag
def get_attr(instance, attr):
    return getattr(instance, attr, "")

@register.simple_tag
def set_attr(instance, attr, value):
    return getattr(instance, attr, value)


@register.simple_tag
def update_dict(context={}, **kwargs):
    context.update(kwargs)
    return context


@register.simple_tag
def parse_into_urlencode(**kwargs):
    return urlencode(kwargs, True)


@register.simple_tag
def add_string(prefix, *args):
    total = prefix
    for element in args:
        total += str(element)
    return total


@register.simple_tag
def is_string(val):
    return isinstance(val, str)

@register.simple_tag
def json_loads(val):
    try:
        return json.loads(val)
    except:
        return {}

@register.simple_tag
def json_dumps(val):
    try:
        return json.dumps(val)
    except:
        return '\{\}'


@register.simple_tag
def apply_func(instance, func_name, *args, **kwargs):
    func = getattr(instance, func_name, None)
    if func is None:
        return ""
    return func(*args, **kwargs)

@register.simple_tag
def get_index(array, index):
    if isinstance(array, dict):
        return array.get(index, None)
    else:
        try:
            return array[index]
        except:
            return None

@register.simple_tag
def get_last_index(array, value):
    pass

@register.simple_tag
def to_camel_case(value):
    return ''.join(x for x in value.title().replace("_", "") if not x.isspace())


@register.simple_tag
def is_on(request, url_name):
    try:
        current_url_name = resolve(request.path_info).url_name
        return current_url_name == url_name
    except Resolver404 as e:
        return False


@register.simple_tag
def get_settings_value(name):
    logger.debug("get_settings_value -> variable -> {} -> {}".format(name, getattr(settings, name, None)))
    return getattr(settings, name, None)


@register.simple_tag
def get_selected_option_value(optgroups):
    for group_name, group_choices, group_index in optgroups:
        if group_choices[0]["selected"]:
            return group_choices[0]["value"]
    return ""


@register.simple_tag
def get_bound_field_context(bound_field):
    widget = bound_field.field.widget

    if bound_field.field.localize:
        widget.is_localized = True

    attrs = {}
    attrs = bound_field.build_widget_attrs(attrs, widget)

    if bound_field.auto_id and 'id' not in widget.attrs:
        attrs.setdefault('id', bound_field.auto_id)

    return widget.get_context(bound_field.html_name, bound_field.value(), attrs)


@register.simple_tag
def get_dict(**kwargs):
    return kwargs


@register.simple_tag
def get_fragments(array, length, fragment_size):
    fragments = []
    actual_length = 0
    while actual_length < length:
        fragments.append(array[actual_length:actual_length+fragment_size])
        actual_length += fragment_size
    return fragments


@register.simple_tag
def assign_variable(variable):
    return variable


@register.simple_tag
def render_to_string(template, **kwargs):
    return render_to_string_or_empty(template, kwargs)


@register.simple_tag
def to_int(val):
    return int(val)


@register.simple_tag
def to_float(val):
    return float(val)

@register.simple_tag
def to_bool(val):
    return bool(val)


@register.simple_tag
def to_list(val):
    return list(val)


@register.simple_tag
def has_key(dict, key):
    return key in dict


@register.simple_tag
def today():
    return timezone.now()


@register.simple_tag
def is_same_date(a, b):
    if (isinstance(a, datetime) or isinstance(a, date)) and (isinstance(b, datetime) or isinstance(b, date)):
        return a.date() == b.date()
    else:
        return False

@register.simple_tag
def get_timedelta(datetime_reference):
    if isinstance(datetime_reference, datetime) or isinstance(datetime_reference, date):
        timedelta = (timezone.now() - datetime_reference)
        timedelta_context = dict()

        days = timedelta.days
        if timedelta.days:
            timedelta_context["days"] = timedelta.days

        minutes = (timedelta.seconds//60)%60
        if minutes:
            timedelta_context["minutes"] = minutes

        hours = timedelta.seconds//3600
        if hours:
            timedelta_context["hours"] = hours

        if timedelta_context.keys() == 0:
            timedelta_context["seconds"] = timedelta.seconds

        return timedelta_context
    else:
        return None

@register.simple_tag
def is_same_year(a, b):
    if isinstance(a, datetime) and isinstance(b, datetime):
        return a.year == b.year
    else:
        return False

@register.simple_tag
def get_class_name(instance):
    return instance.__class__.__name__

@register.simple_tag
def get_domain(url : str):
    return urlparse(url).netloc

@register.simple_tag
def request(method, url, **kwargs):
    try:
        response = requests.request(method, url, **kwargs)
    except:
        logger.exception("Error in the response")
        return {
            "status" : 500,
            "json" : {},
            "content" : ""
        }

    response_content_json = {}
    try:
        response_content_json = response.json()
    except:
        logger.exception("The response isn't a json")

    return {
        "status" : response.status_code,
        "json" : response_content_json,
        "content" : response.content
    }

@register.simple_tag
def build_list(*args):
    return args

@register.simple_tag
def extend_list(elements, *args):
    elements = list(elements) + list(args)
    return elements

@register.simple_tag
def get_index_of_value(array, value, start=0):
    try:
        return array[start:].index(value) + start
    except:
        return None

@register.simple_tag
def get_last_index_of_value(array, value, start=0):
    try:
        return len(array) - 1 - array[start:][::-1].index(value)
    except:
        return None

@register.simple_tag
def subset(array, end, start=0):
    return array[start:end]

@register.simple_tag
def serialize(format, queryset, multi=True, **kwargs):
    try:
        json_list_string = serializers.serialize(format, queryset, **kwargs)
        if multi:
            return json.loads(json_list_string)
        else:
            context = json.loads(json_list_string)
            return context[0]
    except Exception as e:
        if multi:
            return []
        else:
            return {}

@register.simple_tag
def range_diff(index, limit):
    return range(index, limit)


from django import template

register = template.Library()

@register.simple_tag
def truncatesmart(value, limit=80, suffix="..."):
    """
    Truncates a string after a given number of chars keeping whole words.

    Usage:
        {% truncatesmart string }}
        {% truncatesmart string limit=50 }}
    """
    try:
        limit = int(limit)
    # invalid literal for int()
    except ValueError:
        # Fail silently.
        return value

    # Return the string itself if length is smaller or equal to the limit
    if len(value) <= limit:
        return value

    # Break into words and remove the last
    words = value.split(' ')

    current_count = 0
    index = 0

    while (current_count + len(words[index])) < limit:
        current_count += len(words[index])
        if current_count + 1 < limit:
            current_count += 1
        index += 1

    # Join the words and return
    return ' '.join(words[:index]) + suffix
